# coding: utf-8
u"""Модели для хранения данных системы авторизации RBAC."""
from django.contrib.contenttypes.generic import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db import transaction
from django.db.models.fields import FieldDoesNotExist
from django.db.models.signals import post_delete
from django.dispatch import receiver
from m3.db import safe_delete
from south.models import MigrationHistory

from educommon.django.db.mixins.date_interval import ActualObjectsManager
from educommon.django.db.mixins.date_interval import DateIntervalMeta
from educommon.django.db.mixins.date_interval import DateIntervalMixin
from educommon.django.db.utils import model_modifier_metaclass


class _BaseModel(models.Model):

    u"""Базовый класс для моделей приложения."""

    def __unicode__(self):
        return u'{}{{{}}}'.format(self.__class__.__name__, self.pk)

    def full_clean(self, exclude=None):
        self._full_clean_was_called = True

        super(_BaseModel, self).full_clean(exclude=exclude)

    def save(self, *args, **kwargs):
        assert getattr(self, '_full_clean_was_called', False), (
            u'Attempt to save data without validation (model {}.{})'.format(
                self.__class__.__module__, self.__class__.__name__
            )
        )

        for field in self._meta.fields:
            if (isinstance(field, (models.TextField, models.CharField)) and
                    not getattr(self, field.attname)):
                setattr(self, field.attname, None if field.null else u'')

        super(_BaseModel, self).save(*args, **kwargs)

    class Meta:
        abstract = True


class Permission(_BaseModel):

    u"""Разрешение."""

    name = models.CharField(
        u'Имя',
        max_length=100,
        db_index=True,
        unique=True,
    )
    title = models.CharField(
        u'Название',
        max_length=200,
        blank=True, null=True,
    )
    description = models.TextField(
        u'Описание',
        blank=True,
    )

    class Meta:
        verbose_name = u'Разрешение'
        verbose_name_plural = u'Разрешения'

    def __unicode__(self):
        return u'Permission<{}: {}>'.format(self.id, self.name)


class Role(_BaseModel):

    u"""Роль."""

    name = models.CharField(
        u'Название',
        max_length=300,
        db_index=True,
        unique=True,
    )
    description = models.TextField(
        u'Описание',
        blank=True,
    )
    can_be_assigned = models.BooleanField(
        u'Может быть назначена пользователю',
        default=True,
    )
    permissions = models.ManyToManyField(
        Permission,
        related_name='roles',
    )

    class Meta:
        verbose_name = u'Роль'
        verbose_name_plural = u'Роли'

    def __unicode__(self):
        return u'Role<{}: {}>'.format(self.id, self.name)

    @property
    def subroles(self):
        u"""Возвращает все вложенные роли данной роли.

        :rtype: set
        """
        result = set()

        for role_parent in RoleParent.objects.filter(parent_id=self.id):
            result.add(role_parent.role)
            result.update(role_parent.role.subroles)

        return result

    def get_permissions(self):
        u"""Возвращает все разрешения роли, в т.ч. вложенных ролей.

        :rtype: QuerySet
        """
        m2m_model = self._meta.get_field('permissions').rel.through

        roles = set([self]) | self.subroles
        result = Permission.objects.filter(
            pk__in=m2m_model.objects.filter(
                role__in=roles,
            ).distinct('permission').values('permission'),
        )

        return result

    def clean(self):
        errors = {}

        try:
            super(Role, self).clean()
        except ValidationError as error:
            errors = error.update_error_dict(errors)

        if (self.pk and not self.can_be_assigned and
                self.userrole_set.exists()):
            errors.setdefault('can_be_assigned', []).append(
                u'Есть пользователи, которым назначана роль "{}" '.format(
                    self.name
                )
            )

        if errors:
            raise ValidationError(errors)

    @transaction.commit_manually
    def safe_delete(self):
        # M2M-модель, связывающая Role и Permission
        meta = getattr(Role, '_meta')
        m2m_model = meta.get_field('permissions').rel.through

        # Удаление связей разрешений с ролью
        m2m_model.objects.filter(role=self).delete()

        # Удаление связей роли с родительскими ролями
        RoleParent.objects.filter(role=self).delete()

        # Удаление самой роли
        safe_delete(self)

        # safe_delete неправильно работает внутри транзакций, из-за этого
        # при вызове commit() валится IntegrityError, который надо обрабатывать
        # вручную.
        try:
            transaction.commit()
            result = True
        except Exception as error:
            transaction.rollback()

            if error.__class__.__name__ == 'IntegrityError':
                result = False
            else:
                raise

        return result


class RoleParent(models.Model):

    u"""M2M-модель "Вложенная роль"."""

    parent = models.ForeignKey(Role, related_name='+')
    role = models.ForeignKey(Role, related_name='+')

    def clean(self):
        errors = {}

        try:
            super(RoleParent, self).clean()
        except ValidationError as error:
            errors = error.update_error_dict(errors)

        if self.parent.id == self.role.id:
            errors.setdefault('parent', []).append(
                u'Роль не может содержать сама себя'
            )

        # ---------------------------------------------------------------------
        # Проверка отсутствия цикла
        query = RoleParent.objects.all()
        if self.pk:
            query = query.exclude(pk=self.pk)

        def check(target_role, role):
            for role_parent in query.filter(role=role):
                if target_role.id == role_parent.parent_id:
                    raise ValidationError(u'В иерархии ролей обнаружен цикл')
                check(target_role, role_parent.parent)

        try:
            # Проверка, нет ли self.role среди предков self.parent
            check(self.role, self.parent)
        except ValidationError as error:
            errors.setdefault('parent', []).extend(error.messages)
        # ---------------------------------------------------------------------

        if errors:
            raise ValidationError(errors)

    def __unicode__(self):
        return u'RoleParent({} --> {})'.format(
            unicode(self.role), unicode(self.parent)
        )

    class Meta:
        unique_together = ('parent', 'role')


class UserRole(DateIntervalMixin, models.Model):

    u"""M2M-модель "Роль пользователя"."""

    no_intersections_for = ('content_type', 'object_id', 'role')

    __metaclass__ = model_modifier_metaclass(
        DateIntervalMeta,
        date_from=dict(
            verbose_name=u'Действует с',
        ),
        date_to=dict(
            verbose_name=u'по',
        ),
    )

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    user = GenericForeignKey()
    role = models.ForeignKey(
        Role,
        verbose_name=u'Роль',
    )

    objects = models.Manager()
    actual_objects = ActualObjectsManager()

    class Meta:
        verbose_name = u'Роль пользователя'
        verbose_name_plural = u'Роли пользователя'

    def __unicode__(self):
        return u'UserRole({} --> {})'.format(
            unicode(self.user), unicode(self.role),
        )

    def interval_intersected_error_message(self):
        return (
            u'Роль "{}" уже назначена этому пользователю в указанном '
            u'интервале дат.'.format(self.role.name)
        )

    def clean(self):
        errors = {}

        try:
            super(UserRole, self).clean()
        except ValidationError as error:
            errors = error.update_error_dict(errors)

        if not self.role.can_be_assigned:
            errors.setdefault('role', []).append(
                u'Роль "{}" не может быть назначена пользователю'.format(
                    self.role.name
                )
            )

        if errors:
            raise ValidationError(errors)


@receiver(post_delete)
def delete_user_roles(instance, **kwargs):
    u"""Удаление привязки ролей к пользователям при удалении пользователя."""
    if isinstance(instance, MigrationHistory):
        # При накатывании миграций с параметром --delete-ghost-migrations
        # может случиться ситуация, когда для модели UserRole еще нет таблицы.
        # В этом случае ничего не делаем, т.к. иначе будет DatabaseError.
        return

    # Если модель была удалена из Системы, то при накатывании миграций, в
    # которых удаляются записи таких моделей случается AttributeError.
    try:
        content_type = ContentType.objects.get_for_model(instance)
    except AttributeError:
        return

    model_meta = getattr(content_type.model_class(), '_meta')
    try:
        model_meta.get_field_by_name('id')
    except FieldDoesNotExist:
        return

    UserRole.objects.filter(
        content_type=content_type,
        object_id=instance.id,
    ).delete()
